#include "TCPThread.h"

TCPThread::TCPThread( char *name ) : BLooper( name, B_LOW_PRIORITY )
{
	Hostname = "";
	Port = 0;
	socketID = -1;
	socketConnected = false;

	Run();
}

void TCPThread::Connect( char *host, int prt )
{
	Hostname = host;
	Port = prt;

	if( socketConnected )
		Close( );  // Just close the socket and re-connect if already connected
		
	if( !StartConnect( ) )
		Close( );  // There must have been an error spawning the thread!  Give up now!
}

bool TCPThread::StartConnect( )
{
	// Attempt to spawn the socket reader thread
	connectID = spawn_thread( ConnectThreadEntry, "TCPThread::ConnectThread", B_LOW_PRIORITY, this );
	
	// Make sure it got spawned all right
	if( connectID <= 0 )
		return false;
	
	resume_thread( connectID );
	
	return true;
}

int32 TCPThread::ConnectThreadEntry( void *arg )
{
	TCPThread* tcp = (TCPThread*)arg;
	return tcp->ConnectThread( );
}

int32 TCPThread::ConnectThread( )
{
	bool prelim = true;
	bool errorSent = false;
		
	socketID = socket( AF_INET, SOCK_STREAM, 0 );
	
	struct sockaddr_in sa;
	struct hostent *he;
	in_addr ia;	
	int s = sizeof(struct sockaddr_in);
	
	ia.s_addr = INADDR_ANY;
	if( Hostname.Length() != 0 )
	{
		ia.s_addr = inet_addr( Hostname.String() );
		if( ia.s_addr == INADDR_ANY || ia.s_addr == (unsigned long) -1 )
		{
			he = gethostbyname( Hostname.String() );
			if(he != 0)
				ia.s_addr = *(int *)he->h_addr_list[0];
			else
			{
				PostMessage( new BMessage( TCP_ERROR_DNS ) );
				prelim = false;
				errorSent = true;
			}
		}
	}

	sa.sin_family = AF_INET;
	sa.sin_addr = ia;
	sa.sin_port = htons( Port );

	if( prelim )
		if( connect( socketID, (struct sockaddr *) &sa, s ) < 0 )
			prelim = false;
 		else
			prelim = true;

	if( prelim && ( getpeername( socketID, (struct sockaddr *) &sa, &s ) < 0 ) )
		prelim = false;

	if( prelim )
		PostMessage( TCP_CONNECTED );
	else
		if( !errorSent )
			PostMessage( TCP_ERROR_NO_CONNECT );

	return 0;
}


void TCPThread::WaitConnectThread() {
	thread_info info;
	status_t exitcode;
	
	if( get_thread_info( connectID, &info ) != B_BAD_VALUE )
		wait_for_thread( connectID, &exitcode );  // wait for the connect thread to quit
}


bool TCPThread::StartRead( )
{
	// Attempt to spawn the socket reader thread
	readID = spawn_thread( ReadThreadEntry, "TCPThread::ReadThread", B_LOW_PRIORITY, this );
	
	// Make sure it got spawned all right
	if( readID <= 0 )
		return false;
	
	return true;
}

void TCPThread::WakeReadThread() {
	thread_info info;
	
	if( get_thread_info( readID, &info ) == B_NO_ERROR )
		if( info.state == B_THREAD_SUSPENDED )
			resume_thread( readID );
}

void TCPThread::WaitReadThread() {
	thread_info info;
	status_t exitcode;
	
	if( get_thread_info( readID, &info ) != B_BAD_VALUE )
		wait_for_thread( readID, &exitcode );  // wait for the read thread to quit
}

int32 TCPThread::ReadThreadEntry( void *arg )
{
	TCPThread* tcp = (TCPThread*)arg;
	return tcp->ReadThread( );
}


int32 TCPThread::ReadThread( )
{
	bool Run = true;

	while( Run )
	{
		if( socketConnected && ( recv( socketID, NULL, 0, 0 ) >= 0 ) )
		{
			PostMessage( new BMessage( TCP_RECV ) );  // It needs to send a message here instead of just directly calling DataWaiting() otherwise a remote disconnect would go un-noticed!!!!
			suspend_thread( readID );
		}
		else
		{
			if( errno != EINTR )
				Run = false;	// We ran into a problem.  The socket may have disconnected or something.
		}
	}

	Close( true );  // The read thread is dead now, so we better close the socket.  (Or at least try.  It may be closed already and that is why there is nothing to read)
	return 0;
}

ssize_t TCPThread::ReadData( void *buf, size_t size )
{
	long rc;

	rc = recv( socketID, (char *)buf, size, 0 );

	if(rc <= 0)
	{
		Close( ); 	// I'm gonna assume that if there is an error here, we need to stop.
					// It's up to the user to check to make sure that the length is > 0 when this function returns!!!
	}
	
	return rc;
}

ssize_t TCPThread::SendData( const void *buf, size_t size )
{
	ssize_t rc;
	
	rc = send( socketID, (const char *) buf, size, 0 );
/*
	if( rc < 0 )
	{
		//If we end up here, there was a send error.
		//However, I have no idea what to do about it.
		//The user is going to have to check this too.  (Or at least, that would be a good idea)
		//
		////printf( "TCPNet::SendData -> send() ERROR!\n" );
	}
*/
	return rc;
}

void TCPThread::SendStatus( int32 type, int32 error )
{
	BMessage msg( type );
	if( error )
		msg.AddInt32( "TYPE", error );
	msg.AddString( "HOSTNAME", Hostname.String() );
	msg.AddInt32( "PORT", Port );
	
	Invoke( &msg );
}

void TCPThread::Close( bool inReadThread )
{
	if( inReadThread )
		if( LockingThread( ) != -1 )
			return;

	Lock( );

	closesocket( socketID );

	if( !inReadThread )
		WaitReadThread( );  // If we try to kill the read thread while running in the read thread, we hit deadlock.  That would be bad.

	WaitConnectThread( );

	if( socketID >= 0 && socketConnected )
	{
		socketConnected = false;
		socketID = -1;
		SendStatus( TCP_DISCONNECTED );
		Disconnected( );
	}
	Unlock( );
}

void TCPThread::Quit( )
{
	Close( );
	BLooper::Quit( );
}

void TCPThread::MessageReceived( BMessage *msg )
{
	switch( msg->what ) {
		case TCP_ERROR_DNS:
			SendStatus( TCP_ERROR, TCP_ERROR_DNS );
			Error( TCP_ERROR_DNS );
			Close( );
			break;
		case TCP_ERROR_NO_CONNECT:
			SendStatus( TCP_ERROR, TCP_ERROR_NO_CONNECT );
			Error( TCP_ERROR_NO_CONNECT );
			Close( );
			break;
		case TCP_CONNECTED:
			if( StartRead( ) )
			{
				socketConnected = true;
				SendStatus( TCP_CONNECTED );
				Connected( );

				WakeReadThread();  // All right!  Let's open the flood gates....
			}
			else
				Close( );   // If we get here there was a problem spawning the read thread!
			break;
		case TCP_DATA_SEND:
			if( socketConnected )
				SendData( msg->FindString( "DATA" ), msg->FindInt32( "LENGTH" ) );
			break;
		case TCP_RECV:
			if( socketConnected )
			{
				DataWaiting( );
				WakeReadThread( );
			}
			break;
		case TCP_CONNECT:
			Connect( (char *)msg->FindString( "HOSTNAME" ), msg->FindInt32( "PORT" ) );
			break;
		case TCP_DISCONNECT:
			Close( );
			break;
	
		default:
			BLooper::MessageReceived( msg );
			break;
	};
}

void TCPThread::DataWaiting( )
{
	// Blank so that a user can implement it.
 	char Message[256];
 
	ssize_t bytesread = ReadData( Message, 255 );
	
	if( bytesread > 0 )
	{
		Message[bytesread] = '\0';

		BMessage msg( TCP_DATA_RECV );
		msg.AddString( "DATA", Message );
		msg.AddInt32( "LENGTH", strlen( Message ) );
		msg.AddString( "HOSTNAME", Hostname.String() );
		msg.AddInt32( "PORT", Port );
		
		Invoke( &msg );
	}
}
void TCPThread::Connected( )
{
	// Blank so that a user can implement it.
}
void TCPThread::Disconnected( )
{
	// Blank so that a user can implement it.
}
void TCPThread::Error( int32 error )
{
	// Blank so that a user can implement it.
}
void TCPThread::Logout( )
{
	// Blank so that a user can implement it.
}